"Object serialisation in XML for python"

################################################################
#
# A utility library for (de)serializing python objects as
# clean XML, developed for the NeuroML project (www.neuroml.org).
#
#
# It only serialises fields if the name does not start with "_" -
# which skips the internal fields.
#
#
# Fred Howell (fwh@inf.ed.ac.uk)
# v0.0.1 Oct 2005
#
# Known issues:
#   Error reporting is rudimentary - prints to console.
#   Does not yet support the XMLMappingOptions to allow
#     reading/writing weird flavours of XML
#

import sys
from types import *
from sys import maxint, stdout
from xml.dom import pulldom
from os.path import *
import os


#
# A Package is a list of class definitions.
# Fields with a name starting eith "_" are considered
# transient and are ignored by the xml serializer
# (e.g. the _classesIdx field)
#
class Package:    
    def __init__(self, n=""):
        self.name = n
        self.classes=[]
        self._classesIdx={}
        self.desc=""
        
    def addClass(self, cls):
        self._classesIdx[cls.name.lower()]=cls
        self.classes.append(cls)
        cls._package = self

    def getClasses(self):
        return self.classes

    def getClassesMap(self):
        return self._classesIdx

    #
    # Gets a class for a given name, or None if not found.
    # (name is not case sensitive)
    #
    def getClass(self, name):
        if self._classesIdx.has_key(name.lower()):
            return self._classesIdx[name.lower()]
        return None

#
# A Class contains a list of fields.
#
class Class:

    def __init__(self, n=""):
        self.name = n
        self.fields={}
        self.extends = None
        self.keyfield = "name"
        self._package = None
        
    def addField(self, fld):
        self.fields[fld.name]=fld

    #
    # Returns a field of the given name, or None if not found.
    # If this class extends another, finds the superclass
    # from the Package.
    #
    def getField(self, fld):
        if self.fields.has_key(fld):
            return self.fields[fld]
        if self.extends != None:
            if self._package != None:
                base = self._package.getClass(self.extends)
                if base != None:
                    return base.getField(fld)
            else:
                print "error - class "+self.name+" has no package"
        return None
    
    #
    # Returns a suitable field given a type.
    # This is used when parsing XML list members
    # where the field name tag has been omitted.
    #
    def getFieldForType(self, type):
        for f in self.fields:
            # could be "derived from"
            if self.fields[f].elementType == type:
                return self.fields[f]
        return None



#
# A Field has a name and a type.
# Where type is "list" or "ref", the elementType is also set.
# xmlstyle can be showfield | hidefield and governs whether 
# the field name is generated as a tag.
#
class Field:
    
    def __init__(self, n="", t="", et=None, xs=None):
        self.name = n
        self.type = t
        self.elementType = et

        self.xmlstyle = xs
        self.desc = None
        self.units = None
        self.default = None




        
#####################################################################
#
# Mapping options.

#
# Utility class: an iterator for maps.
# Runs over a map returning an object containing
# both key and value.
# If the entries are instances of the given class,
# return each entry in turn; otherwise construct
# an object for each entry based on the
# key and value. This allows a map of string to
# string to be iterated as a sequence of {key,value}
# objects of the appropriate class.
#
class MapIter:

    # d is the map.
    # cls is the name of the class
    def __init__(self, d, cls):        
        self.data = d
        self.list = d.keys()
        self.cls = cls
        self.idx = 0
        self.count = len( self.data.keys() )

    def hasNext(self):
        return self.idx < self.count -1

    def next(self):
        k = self.list[self.idx]
        v = self.data[k]
        self.idx += 1
        if isinstance(v, self.cls):
            return v
        return self.cls(k, v)  




#
# These are global options governing the style
# of XML which is generated / parsed.
# >>> NB : the parser does not yet take these into account! <<<
# Support for these options will be included in a future
# release of the parser.
#
class XMLMappingOptions:
    def __init__(self):        
        self.mappings = {
             "LongNames"      : "useJavaStyle",
             "SimpleFields"   : "useAttributes",
             "ObjectFieldTag" : "useClass",
             "ListFieldTag"   : "hide",
             "ListElementTag" : "useClassName",
             "ClassTag"       : "useClasspath"}
        
    def addMapping(self, m):
        self.mappings[m.key]=m.value
        
    def getMapping(self, key):
        return self.mappings[key]
    
    def getMappingsIterator(self):
        return MapIter(self.mappings, Mapping)

    #
    # Print the options as XML.
    #
    def p(self):
        print "<XMLMappingOptions>"
        for k in self.mappings:
            print '  <Mapping key="'+k+'" value="'+self.mappings[k]+'"/>';
        print "</XMLMappingOptions>"
        

#
# Mapping - a key/value pair.
#
class Mapping:
    def __init__(self, k="", v=""):        
        self.key=k
        self.value=v


#
# Initialise the mappings.
#


mappings = XMLMappingOptions()
# mappings.p()

#
# Create the schema for the mappings.
#
mapPkg = Package("org.neuroml.mapping")
map = Class("Mapping")
map.addField(Field("key", "string"))
map.addField(Field("value", "string"))
mapPkg.addClass(map)
map = Class("XMLMappingOptions")
map.addField(Field("mappings", "list", "Mapping", "hidefield"))
mapPkg.addClass(map)




#####################################################################
#
# serializeObject : converts a python object to XML
#
# obj - the object to serialize
# pkg - the Package instance containing the schema
#
# optional args:
#   ind   - a string indent
#   eltag - the default tagname to use (if None, uses the object class name)
#
# returns - the XML string.
#
# usage: 
# TBD
#
def serializeObject(obj, pkg, ind="", eltag=None):
    tagname = eltag
    
    if tagname==None:
        tagname = obj.__class__.__name__

    # Strings can be vanilla or unicode - return a simple tagged version.
    if isinstance(obj, StringType) or isinstance(obj, UnicodeType):
        return ind+"<"+tagname+">"+obj+"</"+tagname+">\n"

    attrs = ind+"<"+tagname

    # the schemaClass is the Class instance defined by the schema.
    schemaClass = pkg.getClass(obj.__class__.__name__)
    
    if schemaClass == None:
        # error reporting is primitive at present...
        print "Couldn't find pkg cls for "+obj.__class__.__name__
        
    content = ""
    
    # sind is the sub indent for children
    sind = ind + "  "

    # dir(obj) gives an iterator over the fields of the object.
    for membname in dir(obj):

        # skip over any fields starting with "_"
        if membname[0:1]!="_": 
            member = getattr(obj,membname)

            if schemaClass != None:
                schemaField = schemaClass.getField(membname)
            else:
                schemaField = None
            
            prt = True

            # if we've found an object, recurse.
            if type(member) == InstanceType:
                content += serializeObject(member,pkg,sind+"  ")

            # a list or a dictionary
            elif type(member) == ListType or type(member) == DictType:

                # Decide whether or not to include a tag for the field name.
                showfield=True
                if hasattr(schemaField, "xmlstyle"):
                    sty = schemaField.xmlstyle
                    showfield = not (sty == "hidefield")
                    
                if showfield:
                    content += sind+"<"+membname+">\n"
                else:
                    pass

                # Check whether the field has a get<Field>Iterator() method
                method = findMethod(obj.__class__, "get", capinit(membname)+"Iterator")
                if method == None:
                    if type(member) == ListType:
                        for el in member:
                            if issimpletype(schemaField.elementType):
                                # We have a list of simple elements -
                                # use the field name as the tag.
                                eltag = schemaField.name # perhaps singular version...
                                content+=serializeObject(el,pkg,sind+"  ", eltag)
                            else:
                                content+=serializeObject(el,pkg,sind+"  ")
                    else:
                        for (k,el) in member.items():
                            content+=serializeObject(el,pkg,sind+"  ")
                    #content += "xxx end of list"
                else:
                    # use supplied iterator. This is useful for
                    iter = apply(method, [obj])
                    while iter.hasNext():
                        content += serializeObject(iter.next(), pkg, sind+"  ")
                    
                if showfield:
                    content+=sind+"</"+membname+">\n"

            # For strings and unicode fields, serialize as attributes.
            elif isinstance (member,StringType) or isinstance (member,UnicodeType) :
                if (len(member)>0):
                    attrs+=" "+membname+'="'+member+'"'
                    
            # Ints and floats, convert to strings
            elif type(member) == IntType:
                attrs+=" "+membname+'="'+str(member)+'"'
            elif type(member) == FloatType:
                attrs+=" "+membname+'="'+str(member)+'"'
            else:                
                prt = False
                
    if len(content)==0:
        return attrs+"/>\n"
    else:
        if not content.endswith("\n"):
            content += "\n"
        return attrs+">\n"+content+ind+"</"+tagname+">\n"

##################################################################
#
# Utility methods for converting schema into source code.
#
# pygencls(cls) : converts a Class into python source.
#

#
# Generates python source code for a class.
# Can be useful for generating initial cut
# of python class defs from a large schema XML file.
#
# cls - the Class instance
# returns - a string python source code
#
def pygencls(cls):
    ret=""
    ret+= "class "+cls.name
    if hasattr(cls,"extends"):
        ret+="("+cls.extends+"):\n"
    else:
        ret+=":\n"        
    ret+= "    def __init__(self):\n"
    for fn in cls.fields:
        fld = cls.fields[fn]
        if fld.type=="list":
            ret+= "        "+fld.name+"=[]\n"
        elif fld.type=="string":
            ret+= "        "+fld.name+'=""\n'
        else:
            ret+= "        "+fld.name+"=None\n"
    if len(cls.fields)==0:
        ret += "        pass\n"
    return ret

#
# Converts schema types into java types.
#
def javatype(fld):
    if fld.type=="string": return "String"
    if fld.type=="list": return "ArrayList<"+fld.elementType+">"
    return fld.type
    
#
# Generates Java source code from schema.
#
# cls  - the Class object
# pkg  - the Package object
# path - the directory to generate the code into
#
# Fields are generated as private members, using the
# prefix "m_{fieldname}"
#
# get/set methods are also generated.
#
def javagencls(cls, pkg, path):
    # open a file
    file = open(path+cls.name+".java", "w")
        
    file.write("package "+pkg.name+";\n\n")
    file.write("import java.util.*;\n\n")
    file.write( "class "+cls.name)
    if hasattr(cls,"extends") and cls.extends != None:
        file.write( " extends "+cls.extends)
    file.write( " {\n")

    ind = "        ";
    ind2 = ind+ind;
    ind3 = ind2+ind;
    
    # write declarations
    for fn in cls.fields:
        fld = cls.fields[fn]
        jtype = javatype(fld)
        file.write(ind+"private "+jtype+" m_"+fld.name)
        if fld.type=="list":
            file.write(" = new "+jtype+"();\n")
        elif fld.type=="string":
            file.write(' = "";\n') # or default
        else:
            file.write(";\n")

    # write constructor
    file.write(ind+"public "+cls.name+"() {\n");            
    file.write(ind+"}\n")

    # write accessors
    for fn in cls.fields:
        fld = cls.fields[fn]

        jtype = javatype(fld)
        file.write(ind+"public "+jtype+" get"+capinit(fld.name)+"() {\n")
        file.write(ind2+"return m_"+fld.name+";\n")
        file.write(ind+"}\n\n")

        file.write(ind+"public void set"+capinit(fld.name)+"("+jtype+" "+fld.name+") {\n")
        file.write(ind2+"m_"+fld.name+" = "+fld.name+";\n")
        file.write(ind+"}\n\n")

    
    file.write("}\n")
    file.close()
            

#
# Generates Python declarations given a Package.
# returns - the python source code
#
def pygen(pkg):
    ret= "# Python code generated "+pkg.name+"\n"
    for cls in pkg.getClasses():
        ret+=pygencls(cls)
    return ret

#
# Generates Java class definitions for a Package.
#
# Output is in a directory defined by the
# package name.
#
def javagen(pkg):
    path = pkg.name.split(".")
    up=""
    p = ""
    for s in path:
        p+=s+"/"
        if not exists(p):
            os.mkdir(p)
    for cls in pkg.getClasses():
        javagencls(cls, pkg, p)



###############################################################
#        
# Define python classes for schema.
schemaPkg = Package("org.neuroml.schema")

cls = Class("Package")
schemaPkg.addClass(cls)
cls.addField( Field("name", "string") )
cls.addField( Field("classes", "list", "Class", "hidefield") )

cls2 = Class("Class")
cls2.addField( Field("name", "string") )
cls2.addField( Field("fields", "list", "Field", "hidefield") )
schemaPkg.addClass(cls2)

cls2 = Class("Field")
cls2.addField( Field("name", "string") )
cls2.addField( Field("type", "string") )
schemaPkg.addClass(cls2)



##
# PP is a simple pull parser wrapper around the built in
# pulldom layer.
# The extra feature it adds is the ability to
# putback() an event.
#
class PP:

    #
    # Initialises the parser from a file
    #
    def __init__(self, file):
        self.events = pulldom.parse(file)
        self.event = None
        self.node = None
        self.pb = False # putback buffer contains an event?

    #
    # Returns the next event from the parser.
    #
    def next(self):
        if self.pb:
            self.pb = False
            return self.event
        (self.event, self.node) = self.events.next()
        # skip whitespace entries
        textfound = False
        while (self.event==pulldom.CHARACTERS and not textfound):
            if len(self.text()) == 0:
                (self.event, self.node) = self.events.next()
            else:
                textfound = True
        return self.event

    #
    # Putback an event.
    #
    def putback(self):
        self.pb = True

    #
    # Rest of methods are defined by the pulldom interface.
    #
    def tag(self):
        return self.node.tagName
    def attr(self):
        return self.node.attributes
    def type(self):
        return self.event
    def require(self, type):
        return self.event == type
    def text(self):
        return self.node.data.strip()
    
#
# Utility method : Splits a list in format "aaa|bbb|ccc" into [aaa,bbb,ccc]
#
def splitList(str):
    return str.split('|')


#
# Splits a list in format key: val (aaa|bbb|ccc) val2 (fff|ggg|hhh)"
# returns (key, dict)
# where dict is a map from value to list (string)
#
def splitKeyList(str):
    (key, rest) = str.split(':')
    key = key.strip()
    entries = rest.split(')')
    ret = {}
    for e in entries:
        (val, options) = e.split("(")
        optlist = options.split("|")
        val = val.strip()
        ret[val] = optlist    
    return (key, ret)


########################################################
#
# Parses XML, returning a string or an object.
# Calls itself recursively to parse entire object tree.
#
# pp     : a PP instance
# ns     : a locals() with the classes defined.
# pkg    : an instance of Package with the schema Classes.
# simple : a flag - if true, expecting a simple entry (no subobjects)
#
# returns (obj, schemaClass)
#
def parseXML(pp, ns, pkg, simple=False):

    # simple : looking for <tag>cvalue</tag>
    if simple:
        val = fetchsimple(pp)
        return (val, "string")

    # print "---START "+pp.tag()

    if not pp.require(pulldom.START_ELEMENT):
        print "error - expecting start element, got:"
        print pp.type()
        
    scratch = None
    anyAttrs = False
    anyChildren = False

    # Check if tag found in pkg schema
    if pp.tag().lower() in pkg.getClassesMap():
        schemaClass = pkg.getClass( pp.tag() )

        if not schemaClass.name in ns: # make it...
            print schemaClass.name+" not found in locals"
            exec ('class %s: pass' % schemaClass.name)
        
        codeClass = ns[schemaClass.name]
        obj = codeClass()
    else:
        schemaClass = None
        print "class "+pp.tag()+" not found"

    # set attributes.
    ats = pp.attr()
    for k in ats.keys():
        setattr(obj, k, str(ats[k].value))
        anyAttrs = True

    # Fetch next tag until hit the </> end element.
    while pp.next() != pulldom.END_ELEMENT:
        if pp.type() == pulldom.CHARACTERS:
            # scratch accumulates text - which may arrive in multiple CHARACTERS events.
            if scratch==None: scratch=""
            scratch += str(pp.text())
        elif pp.type() == pulldom.START_ELEMENT:
            anyChildren=True

            fld = pp.tag()

            # clsfld is the Field object defined in the schema
            clsfld = None
            if schemaClass != None:
                clsfld = schemaClass.getField(fld)
            
            inlinelist = False
            
            if schemaClass != None and clsfld == None:
                # Lookup a suitable list for this type.
                clsfld = schemaClass.getFieldForType(fld)
                if clsfld != None:
                    inlinelist = True
                    # could possibly be an Object field.
            elif schemaClass != None:
                if clsfld.type=="list" and clsfld.xmlstyle=="hidefield":
                    inlinelist = True
                    
            if clsfld == None:
                # OK, this is an element of a list...
                # which list?
                # print "making field "+fld+ " in obj "+obj.__class__.__name__
                clsfld = Field(fld, "string")
                if schemaClass != None:
                    schemaClass.addField(clsfld)

            # We think this field is a list.
            if clsfld.type=="list":
                if hasattr(obj, clsfld.name):
                    fldobj = getattr(obj, clsfld.name) # instance object (may be None)
                else:
                    fldobj = None

                isdict = False
                if isinstance(fldobj,ListType):
                    listobj = fldobj
                elif isinstance(fldobj, DictType):
                    listobj = fldobj
                    isdict = True
                else:
                    listobj = []  ## get the instance object
                    setattr(obj, clsfld.name, listobj)
                    
                # look for an "append" method for this field,
                # which allows user code to be called during
                # parsing a list of things.
                method = findAppendMethod(codeClass, clsfld.name, clsfld.elementType)
                
                simplelist = issimpletype(clsfld.elementType)
                
                # print "START LIST ---" +clsfld.name+" "+str(inlinelist)
                cnt = 0
                if inlinelist:
                    atend = False
                    while not atend:
                        # check if next one is same class...
                        atend = pp.tag() != fld
                        if not atend:
                            (elem, elemClass) = parseXML(pp, ns, pkg, simplelist)
                            
                            addListItem(obj, elem, elemClass, listobj, method, isdict)
                            cnt += 1
                            if cnt % 100 == 0:
                                # debug messages every 100 items.
                                print "list entry "+clsfld.name+"   "+str(cnt)
                        
                            pp.next()
                            atend = pp.type() != pulldom.START_ELEMENT
                        if atend:
                            pp.putback()
                else:
                    while pp.next() == pulldom.START_ELEMENT:
                        (elem, elemClass) = parseXML(pp,ns, pkg, simplelist)
                        cnt += 1
                        if cnt % 100 == 0:
                            # debug messages every 100 items.
                            print "list entry "+clsfld.name+"   "+str(cnt)


                        addListItem(obj, elem, elemClass, listobj, method, isdict)
                # print "END LIST ---"+clsfld.name                    
                    
            else:
                # print "found a simple type perhaps "+pp.tag()
                # could be a simple type...
                # do a sniff...
                foundSimple = False
                fldTag = pp.tag()
                foundText=None
                sniff = pp.next()
                if sniff==pulldom.CHARACTERS:
                    foundSimple = True
                    foundText = pp.text()
                    while pp.next() == pulldom.CHARACTERS:
                        foundText+=pp.text()
                    if pp.type() != pulldom.END_ELEMENT:
                        print "Error - found mixed content "+fldTag+" "+pp.type()

                elif sniff==pulldom.END_ELEMENT:
                    foundSimple=True
                    
                    # expectingSimple = "string|int|double".find(clsfld.type) != -1:
                if foundSimple:
                    if foundText != None:
                        # print "setting attribute "+fldTag+" to "+foundText
                        setattr(obj, fldTag, foundText)
                else:
                    # try and load an object instead
                    (subobj, subcls) = parseXML(pp,ns,pkg)
                    setattr(obj, pp.tag(), subobj)
                    # print "found a "+subobj.__class__.__name__+" "+pp.type()+" "+pp.tag()
                    pp.next() # to skip over end of field.

        else:
            print "unknown type "+pp.tag()

    #print "---END "+pp.tag()

    if (not anyChildren) and (not anyAttrs):
        return (scratch, "string")
    #print "made an object "
    #print obj
    return (obj, schemaClass)


#
# Check if a type is simple.
#
def issimpletype(str):
    return "string|int|double".find(str) != -1

#
# Fetch a simple <tag>value</tag> from the parser.
# pp - the PP instance
# returns - a string.
#
def fetchsimple(pp):
    foundText=None
    sniff = pp.next()
    
    if sniff==pulldom.CHARACTERS:
        foundSimple = True
        foundText = pp.text()
        while pp.next()==pulldom.CHARACTERS:
            fountText+=pp.text()
        if pp.type() != pulldom.END_ELEMENT:
            print "Error - found mixed content "+fldTag
    elif sniff==pulldom.END_ELEMENT:
        foundSimple=True
        foundText=""
    if foundSimple:
        return foundText
    pp.putback()
    return None


#
# Adds a list or dict item, using append method if one supplied.
#
def addListItem(obj, elem, elemClass, listobj, method, isdict):
    if method != None: #None:
        #print "applying append method "
        #print method
        apply(method, (obj, elem))
    elif isdict:
        # find the key field...
        keyfield = elemClass.keyfield
        if keyfield==None: keyfield="name"
        keyval = getattr(elem, keyfield) # *** could return None
        listobj[keyval] = elem
    else:
        listobj.append(elem)

#
# Look for a method named <pfix><Name>
#
def findMethod(cls, pfix, name):
    fullName = pfix + name
    # print "looking for method "+fullName
    if hasattr(cls, fullName):
        return getattr(cls, fullName)
    return None

#
# Finds an add<Type> ot appendTo<Listfld> method
#
def findAppendMethod(cls, listfld, type):
    appendMethod = findMethod(cls, "appendTo", capinit(listfld))

    if appendMethod != None:
        return appendMethod

    addMethod = findMethod(cls, "add", type)
    if addMethod != None:
        return addMethod
    
    return None

#
# Returns a Capitalized version of a string.
#
def capinit(str):
    return str[0].upper()+str[1:]

####################################################
#
# Parses an xml file.
#
# fn  - filename
# ns  - namespace. This is the locals() variable where the python class defs live.
# pkg - the schema Package object, perhaps loaded from a file, or created by hand.
#
# returns - a python object
#
def parseXMLFile(fn, ns, pkg):
    pp = PP(open(fn))
    scratch = None

    while pp.next() != pulldom.START_ELEMENT:
        pass
    (tree, cls) = parseXML(pp, ns, pkg)
    return tree

schemaNs = locals()


#
# Parses a schema file
# returns - a Package object
#
def parseSchemaFile(fn):
    return parseXMLFile(fn, schemaNs, schemaPkg)

#
# Parses a Map file [not used by parser yet]
# returns - an XMLMappingOptions object
def parseMapFile(fn):
    return parseXMLFile(fn, schemaNs, mapPkg)


